
module arhuman;

import stream;

// Les diffrentes classes de caractre que
// l'on peut croiser dans cette grammaire

enum CharType {

	// caractres spciaux
	NEWLINE,	     // saut de ligne
	EOF,		     // fin de fichier
	WHITE,		     // \r \t ' '
	DIESE,		     // # commentaire

	// caractres de structure
	OPENING_PARENTHESIS, // (
	CLOSING_PARENTHESIS, // )
	OPENING_BRACKET,     // {
	CLOSING_BRACKET,     // }
	SEMICOLON,	     // ;
	QUOTE,		     // "

	// oprateurs
	PLUS,		     // +
	MINUS,		     // -
	DIV,		     // /
	MUL,		     // *
	GT,		     // >
	LT,		     // <
	EQUAL,		      // =
	MODULO, 	      // %

	// identifiants & litteraux
	LETTER, 	     // [a..zA..Z]
	DIGIT,		     // [0..9]
	DOT,		     // .


	OTHER // le reste
}

CharType[] createCharacterTable() {

	// on cr le tableau
	CharType[] ct = new CharType[256];

	// on l'initialise
	ct[] = CharType.OTHER;

	// Lettre minuscules
	ct['a'..'z'+1] = CharType.LETTER;

	// Lettre majuscules
	ct['A'..'Z'+1] = CharType.LETTER;

	// Les digits
	ct['0'..'9'+1] = CharType.DIGIT;
	ct['.'] = CharType.DOT;

	ct["\n"] = CharType.NEWLINE;
	ct["\r"] = CharType.WHITE;
	ct["\t"] = CharType.WHITE;
	ct[' '] = CharType.WHITE;
	ct['#'] = CharType.DIESE;

	ct['('] = CharType.OPENING_PARENTHESIS;
	ct[')'] = CharType.CLOSING_PARENTHESIS;
	ct['{'] = CharType.OPENING_BRACKET;
	ct['}'] = CharType.CLOSING_BRACKET;
	ct[';'] = CharType.SEMICOLON;
	ct['"'] = CharType.QUOTE;

	ct['+'] = CharType.PLUS;
	ct['-'] = CharType.MINUS;
	ct['/'] = CharType.DIV;
	ct['*'] = CharType.MUL;
	ct['>'] = CharType.GT;
	ct['<'] = CharType.LT;
	ct['='] = CharType.EQUAL;
	ct['%'] = CharType.MODULO;

	return ct;
}

// Les type de tokens maintenant
enum TokenType {

	PRINT,	   // print
	IF,	   // if
	WHILE,	   // while
	DEFINE,    // define
	PLUS,	   // +
	MINUS,	   // -
	DIV,	   // /
	MUL,	   // *
	MODULO,    // %
	GT,	   // >
	LT,	   // <
	GTOE,	   // >=
	LTOE,	   // <=
	EQUAL,	   // ==
	NEQUAL,    // <>
	ASSIGN,    // =
	OPENING_PARENTHESIS,   // (
	CLOSING_PARENTHESIS,   // )
	OPENING_BRACKET, // {
	CLOSING_BRACKET, // }
	LITTERAL,	 // 1.0, 1231, 21321321.211222
	IDENTIFIER,	 // toto, a1, hey
	STRING, 	 // "bla bla bla"
	SEMICOLON,	 // ;
	EOF		 // fin de fichier

}

struct Token {
	TokenType tt;
	char[] representation;
	int line;
	int pos;
}

enum LexerState {
	DEFAULT,
	IN_COMMENT,
	IN_STRING,
	IN_LITTERAL,
	IN_IDENTIFIER,
	MAYBE_GTOE,
	MAYBE_LTOE_OR_NEQUAL,
	MAYBE_EQUAL
}

class Lexer {

	static CharType[] ctable;
	static this() {
		ctable = createCharacterTable();
	}

	LexerState ls = LexerState.DEFAULT;
	CharType ct;
	char c;
	Stream input;
	int line = 1,pos = 1;

	this( Stream input ) {
		this.input = input;
		step();
	}

	private void step() {
		if( input.eof() ) {
			c = "\0";
			ct = CharType.EOF;
		} else {
			c = input.getc();
			ct = ctable[c];
		}
		if( ct == CharType.NEWLINE ) {
			pos = 1;
			line++;
		} else {
			pos++;
		}
	}


	public Token nextToken() {

		Token t;

		char[] buffer;

		while(true) {
			switch(ls) {

			case LexerState.DEFAULT:

				t.line = line; // save line
				t.pos  = pos;  // save pos

				switch(ct) {

				case CharType.DIESE: // comment
					ls = LexerState.IN_COMMENT;
					break;

				case CharType.NEWLINE:
				case CharType.WHITE:
					step(); // skip
					break;

				case CharType.LETTER:
					buffer ~= c;
					step();
					ls = LexerState.IN_IDENTIFIER;
					break;

				case CharType.DIGIT:
					buffer ~= c;
					step();
					ls = LexerState.IN_LITTERAL;
					break;

				case CharType.GT:
					buffer ~= c;
					step();
					ls = LexerState.MAYBE_GTOE;
					break;

				case CharType.LT:
					buffer ~= c;
					step();
					ls = LexerState.MAYBE_LTOE_OR_NEQUAL;
					break;

				case CharType.EQUAL:
					buffer ~= c;
					step();
					ls = LexerState.MAYBE_EQUAL;
					break;

				case CharType.PLUS:
					t.tt = TokenType.PLUS;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.MINUS:
					t.tt = TokenType.MINUS;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.DIV:
					t.tt = TokenType.DIV;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.MUL:
					t.tt = TokenType.MUL;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.MODULO:
					t.tt = TokenType.MODULO;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.OPENING_PARENTHESIS:
					t.tt = TokenType.OPENING_PARENTHESIS;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.CLOSING_PARENTHESIS:
					t.tt = TokenType.CLOSING_PARENTHESIS;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.OPENING_BRACKET:
					t.tt = TokenType.OPENING_BRACKET;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.CLOSING_BRACKET:
					t.tt = TokenType.CLOSING_BRACKET;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.SEMICOLON:
					t.tt = TokenType.SEMICOLON;
					t.representation ~= c;
					step();
					return t;
					break;

				case CharType.QUOTE:
					t.tt = TokenType.STRING;
					step();
					ls = LexerState.IN_STRING;
					break;

				case CharType.EOF:
					t.tt = TokenType.EOF;
					t.representation = "EOF";
					return t;
					break;

				default:
					printf( "Unexpected character line %d, char %d\n", line,pos );
					throw new Error( "Unexpected character" );
					break;
				}
				break;

			case LexerState.IN_COMMENT:
				switch(ct) {

				case CharType.NEWLINE: // end comment
					ls = LexerState.DEFAULT;
					step();
					break;

				case CharType.EOF:
					ls = LexerState.DEFAULT;
					break;

				default:
					step(); // skip
					break;
				}
				break;


			case LexerState.IN_STRING:
				switch(ct) {

				case CharType.QUOTE: // string
					ls = LexerState.DEFAULT;
					step();
					return t;
					break;

				case CharType.NEWLINE:
				case CharType.EOF:
					printf( "Expected end of string line %d, char %d\n", line,pos );
					throw new Error( "fatal error" );
					break;

				default:

					t.representation ~= c;
					step();
					break;
				}
				break;


			case LexerState.IN_IDENTIFIER:
				switch(ct) {

				case CharType.LETTER:
				case CharType.DIGIT:
					buffer ~= c;
					step();
					break;

				default:
					switch(buffer) {
					case "print": t.tt = TokenType.PRINT; break;
					case "if": t.tt = TokenType.IF; break;
					case "while": t.tt = TokenType.WHILE; break;
					case "define": t.tt = TokenType.DEFINE; break;
					default: t.tt = TokenType.IDENTIFIER; break;
					}
					t.representation ~= buffer;
					buffer.length = 0;
					ls = LexerState.DEFAULT;
					return t;
					break;

				}
				break;


			case LexerState.IN_LITTERAL:
				switch(ct) {

				case CharType.DOT:
				case CharType.DIGIT:
					buffer ~= c;
					step();
					break;

				default:
					t.tt = TokenType.LITTERAL;
					t.representation ~= buffer;
					buffer.length = 0;
					ls = LexerState.DEFAULT;
					return t;
					break;

				}
				break;

			case LexerState.MAYBE_GTOE:
				switch(ct) {

				case CharType.EQUAL:
					buffer ~= c;
					t.tt = TokenType.GTOE;
					t.representation = buffer;
					step();
					ls = LexerState.DEFAULT;
					buffer.length = 0;
					return t;
					break;

				default:
					t.tt = TokenType.GT;
					t.representation = buffer;
					buffer.length = 0;
					ls = LexerState.DEFAULT;
					return t;
					break;

				}
				break;

			case LexerState.MAYBE_LTOE_OR_NEQUAL:
				switch(ct) {

				case CharType.EQUAL:
					buffer ~= c;
					t.tt = TokenType.LTOE;
					t.representation = buffer;
					step();
					ls = LexerState.DEFAULT;
					buffer.length = 0;
					return t;
					break;

				case CharType.GT:
					buffer ~= c;
					t.tt = TokenType.NEQUAL;
					t.representation = buffer;
					step();
					ls = LexerState.DEFAULT;
					buffer.length = 0;
					return t;
					break;

				default:
					t.tt = TokenType.LT;
					t.representation = buffer;
					buffer.length = 0;
					ls = LexerState.DEFAULT;
					return t;
					break;

				}
				break;


			case LexerState.MAYBE_EQUAL:
				switch(ct) {

				case CharType.EQUAL:
					buffer ~= c;
					t.tt = TokenType.EQUAL;
					t.representation = buffer;
					step();
					ls = LexerState.DEFAULT;
					buffer.length = 0;
					return t;
					break;

				default:
					t.tt = TokenType.ASSIGN;
					t.representation = buffer;
					buffer.length = 0;
					ls = LexerState.DEFAULT;
					return t;
					break;

				}
				break;
			}
		}
	}
}

class SymbolTable {

	private SymbolTable superTable = null;
	private float[char[]] table;

	this( SymbolTable st ) {
		this.superTable = st;
	}

	bit isDirectlyDefined( char[] identifier ) {
		return identifier in table;
	}

	bit isDefined( char[] identifier ) {
		if( identifier in table ) {
			return true;
		} else if( superTable != null ) {
			return superTable.isDefined(identifier);
		}
		return false;
	}

	void define( char[] identifier ) {
		table[identifier] = 0.0;
	}

	void set( char[] identifier , float value ) {
		if( isDirectlyDefined( identifier )) {
			table[identifier] = value;
		} else {
			superTable.set( identifier , value );
		}
	}

	float get( char[] identifier ) {
		switch(identifier) {
		case "true": return 1.0;
		case "false": return 0.0;
		case "pi": return 3.141592;
		case "e" : return 2.718281;
		case "or" : return 1.618;

		default:
		if( isDirectlyDefined( identifier ) ) {
			return table[identifier];
		} else {
			return superTable.get( identifier );
		}
		}
	}

	void flush() {
		char[][] keys = table.keys;
		for( int i = 0 ; i < keys.length ; i++ )
			delete table[keys[i]];
	}
}

class Node {}

class StatementNode : Node {
	public void execute() {};
}

class DefineNode : StatementNode {
	public {
		SymbolTable st;
		Token define;
		Token identifier;
		Token semicolon;
	}
	public void execute() {
		char[] i = identifier.representation;
		if( st.isDirectlyDefined(i) ) {
			printf( "(line %d): " ~ i ~ " already defined\n",
				define.line );
			throw new Error( "fatal error" );
		}
		st.define( i );
	}
}

class IfNode : StatementNode {
	public {
		Token _if;
		Token opening;
		ExpressionNode expression;
		Token closing;
		BlockNode block;
	}
	public void execute() {
		float e = expression.evaluate();
		if( e != 0.0 )
			block.execute();
	}
}

class WhileNode : StatementNode {
	public {
		Token _while;
		Token opening;
		ExpressionNode expression;
		Token closing;
		BlockNode block;
	}
	public void execute() {
		while( expression.evaluate() != 0.0 )
			block.execute();
	}
}

class AssignNode : StatementNode {
	public {
		SymbolTable st;
		Token identifier;
		Token assign;
		ExpressionNode expression;
		Token semicolon;
	}
	public void execute() {
		char[] i = identifier.representation;
		if( ! st.isDefined( i )) {
			printf( "(line %d): " ~ i ~ " is not defined\n",
				identifier.line );
			throw new Error( "fatal error" );
		}
		st.set( i , expression.evaluate() );
	}

}

class PrintNode : StatementNode {
	public {
		bit isString = false;
		Token print;
		Token opening;
		ExpressionNode expression;
		Token string;
		Token closing;
		Token semicolon;
	}
	public void execute() {
		if( isString ) {
		printf( "(line %d): " ~ string.representation ~"\n",
			print.line );
		} else {
		printf("(line %d): %.10g\n" ,
			print.line,
			expression.evaluate() );
		}
	}
}

class ExpressionNode : Node {
	public float evaluate() { return 0.0; }
}

class IdentifierNode : ExpressionNode {
	public {
		SymbolTable st;
		Token identifier;
	}

	public float evaluate() {
		return st.get(identifier.representation);
	}
}

class LitteralNode : ExpressionNode {
	public {
		Token litteral;
	}

	public float evaluate() {
		return atof( litteral.representation );
	}
}

class OperationNode : ExpressionNode {
	public {
		Token opening;
		ExpressionNode expression1;
		Token operator;
		ExpressionNode expression2;
		Token closing;
	}

	public float evaluate() {
		float e1 = expression1.evaluate();
		float e2 = expression2.evaluate();

	switch( operator.tt ) {

	case TokenType.PLUS:
		return e1 + e2;
	case TokenType.MINUS:
		return e1 - e2;
	case TokenType.DIV:
		if( e2 == 0.0 ) {
			printf( "(line %d): division by zero\n",
				opening.line );
			throw new Error("fatal error");
		}
		return e1 / e2;
	case TokenType.MUL:
		return e1 * e2;
	case TokenType.MODULO:
		return e1 % e2;
	case TokenType.GT:
		return ( e1 > e2 ) ? 1.0 : 0.0;
	case TokenType.LT:
		return ( e1 < e2 ) ? 1.0 : 0.0;
	case TokenType.GTOE:
		return ( e1 >= e2 ) ? 1.0 : 0.0;
	case TokenType.LTOE:
		return ( e1 <= e2 ) ? 1.0 : 0.0;
	case TokenType.EQUAL:
		return ( e1 == e2 ) ? 1.0 : 0.0;
	case TokenType.NEQUAL:
		return ( e1 != e2 ) ? 1.0 : 0.0;

	}
	}
}

class NegateNode : ExpressionNode {
	public {
		Token minus;
		ExpressionNode expression;
	}

	public float evaluate() {
		return - expression.evaluate();
	}
}

class BlockNode : Node {
	public void execute() {}
}

class SingleNode : BlockNode {
	public {
		SymbolTable st;
		StatementNode statement;
	}
	public void execute() {
		statement.execute();
		st.flush();
	}
}

class MultipleNode : BlockNode {
	public {
		SymbolTable st;
		Token opening;
		StatementNode[] statements;
		Token closing;
	}
	public void execute() {
		for( int i = 0 ; i < statements.length ; i++ )
			statements[i].execute();
		st.flush();

	}
}

class AxiomNode : BlockNode {
	public {
		SymbolTable st;
		StatementNode[] statements;
	}
	public void execute() {
		for( int i = 0 ; i < statements.length ; i++ )
			statements[i].execute();
		st.flush();

	}
}



class Parser {

	Lexer lexer;
	Token current;

	this( Lexer l ) {
		this.lexer = l;
		step();
	}

	private void error( char[] message ) {
		printf( "(line %d): " ~ message ~ "\n",
			current.line );
		throw new Error("fatal error");
	}

	private void step() {
		current = lexer.nextToken();
	}

	private bit isOperator( TokenType t ) {

	switch( t ) {
	case TokenType.PLUS:
	case TokenType.MINUS:
	case TokenType.DIV:
	case TokenType.MUL:
	case TokenType.GT:
	case TokenType.LT:
	case TokenType.GTOE:
	case TokenType.LTOE:
	case TokenType.EQUAL:
	case TokenType.NEQUAL:
	case TokenType.MODULO:
		return true;
		break;

	default:
		return false;
		break;

	}
	}

	private ExpressionNode getExpression(SymbolTable st) {

		switch(current.tt) {

		case TokenType.IDENTIFIER: {
			IdentifierNode node = new IdentifierNode();
			node.st = st;
			node.identifier = current;
			step();
			return node;
			}

		case TokenType.LITTERAL: {
			LitteralNode node = new LitteralNode();
			node.litteral = current;
			step();
			return node;
			}

		case TokenType.OPENING_PARENTHESIS: {
			OperationNode node = new OperationNode();
			node.opening = current;
			step();
			node.expression1 = getExpression(st);
			if( ! isOperator( current.tt ))
				error( "Operator expected" );
			node.operator = current;
			step();
			node.expression2 = getExpression(st);
			if( current.tt != TokenType.CLOSING_PARENTHESIS )
				error( "')' expected" );
			node.closing = current;
			step();
			return node;
			}


		case TokenType.MINUS: {
			NegateNode node = new NegateNode();
			node.minus = current;
			step();
			node.expression = getExpression(st);
			return node;
			}

		default: {
			error( "Expression expected" );
			}
		}

	}

	private StatementNode getStatement(SymbolTable st) {

		switch(current.tt) {

		case TokenType.DEFINE: {
			DefineNode node = new DefineNode();
			node.st = st;
			node.define = current;
			step();
			if( current.tt != TokenType.IDENTIFIER )
				error( "Identifier expected" );
			node.identifier = current;
			step();
			if( current.tt != TokenType.SEMICOLON )
				error( "';' expected" );
			node.semicolon = current;
			step();
			return node;
			}

		case TokenType.IF: {
			IfNode node = new IfNode();
			node._if = current;
			step();
			if( current.tt != TokenType.OPENING_PARENTHESIS )
				error( "'(' expected" );
			node.opening = current;
			step();
			node.expression = getExpression(st);
			if( current.tt != TokenType.CLOSING_PARENTHESIS )
				error( "')' expected" );
			node.closing = current;
			step();
			node.block = getBlock(st);
			return node;
			}

		case TokenType.WHILE: {
			WhileNode node = new WhileNode();
			node._while = current;
			step();
			if( current.tt != TokenType.OPENING_PARENTHESIS )
				error( "'(' expected" );
			node.opening = current;
			step();
			node.expression = getExpression(st);
			if( current.tt != TokenType.CLOSING_PARENTHESIS )
				error( "')' expected" );
			node.closing = current;
			step();
			node.block = getBlock(st);
			return node;
			}

		case TokenType.IDENTIFIER: {
			AssignNode node = new AssignNode();
			node.st = st;
			node.identifier = current;
			step();
			if( current.tt != TokenType.ASSIGN )
				error( "'=' expected" );
			node.assign = current;
			step();
			node.expression = getExpression(st);
			if( current.tt != TokenType.SEMICOLON )
				error( "';' expected" );
			node.semicolon = current;
			step();
			return node;
			}

		case TokenType.PRINT: {
			PrintNode node = new PrintNode();
			node.print = current;
			step();
			if( current.tt != TokenType.OPENING_PARENTHESIS )
				error("'(' expected" );
			node.opening = current;
			step();
			if( current.tt == TokenType.STRING ) {
				node.string = current;
				node.isString = true;
				step();
			} else {
				node.expression = getExpression(st);
				node.isString = false;
			}
			if( current.tt != TokenType.CLOSING_PARENTHESIS )
				error( "')' expected" );
			node.closing = current;
			step();
			if( current.tt != TokenType.SEMICOLON )
				error( "';' expected" );
			node.semicolon = current;
			step();
			return node;
			}

		default: {
			error( "Statement expected" );
			return null;
			}
		}
	}

	private BlockNode getBlock(SymbolTable symbol ) {

		SymbolTable st = new SymbolTable( symbol );

		switch(current.tt) {

		case TokenType.OPENING_BRACKET : {
			MultipleNode node = new MultipleNode();
			node.st = st;
			node.opening = current;
			step();
			while( current.tt != TokenType.CLOSING_BRACKET )
				node.statements ~= getStatement(st);
			node.closing = current;
			step();
			return node;
			}

		default: {
			SingleNode node = new SingleNode();
			node.st = st;
			node.statement = getStatement(st);
			return node;
			}
		}
	}

	public AxiomNode parse() {
		AxiomNode node = new AxiomNode();
		SymbolTable st = new SymbolTable(null);
		node.st = st;
		while(current.tt != TokenType.EOF) {
			node.statements ~= getStatement(st);
		}
		return node;
	}

}

int main( char[][] args ) {

	if( args.length != 2 ) {
		printf( "usage: arhuman arhuman.file\n" );
		return 1;
	}

	File f = new File( args[1] , FileMode.In );

	if( ! f.readable ) {
		printf( "bad file : " ~ args[1] );
		return 1;
	}

	Lexer l = new Lexer( f );
	Parser p = new Parser( l );

	AxiomNode node = p.parse();

	f.close();

	node.execute();

	return 0;

}
